#
#  engine
#
#  Playout Daemon for autoradio project
#
#
#  Copyright (C) 2017-2018 Gottfried Gaisbauer <gottfried.gaisbauer@servus.at>
#
#  This file is part of engine.
#
#  engine is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  engine is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with engine. If not, see <http://www.gnu.org/licenses/>.
#

import os
import sys
import threading
import json
import queue
import traceback
import urllib
import logging

from mutagen.flac import FLAC
from datetime import datetime, timedelta

from libraries.database.broadcasts import Schedule, Playlist, PlaylistEntry, PlaylistEntryMetaData
from libraries.enum.auraenumerations import ScheduleEntryType
from modules.communication.redis.messenger import RedisMessenger
from modules.scheduling.calender_fetcher import CalendarFetcher


class AuraCalendarService(threading.Thread):
    messenger = None
    until = ""
    playlistdir = ""
    xmlplaylist = range(0)

    queue = None
    config = None
    debug = False
    _stop_event = None
    logger = None
    fetched_schedule_data = None
    url = dict()
    data = dict()
    calendar_fetcher = None

    """
    Fetching playlist data, write it into the database and notify service
    """
    def __init__(self, config, datefrom="", dateto=""):
        threading.Thread.__init__(self)
        self.config = config
        self.messenger = RedisMessenger(config)
        self.logger = logging.getLogger("AuraEngine")

        self.messenger.set_channel("aura")
        self.messenger.set_section("calendar")

        self.datefrom = str(datefrom)
        self.dateto = dateto

        self.queue = queue.Queue()

        self._stop_event = threading.Event()

        self.__set_url__("calendar")
        self.__set_url__("importer")

        self.calendar_fetcher = CalendarFetcher(config)

    # ------------------------------------------------------------------------------------------ #
    def set_date_from(self, date):
        self.datefrom = str(date).replace(" ", "T")

    # ------------------------------------------------------------------------------------------ #
    def set_date_to(self, date):
        self.dateto = str(date).replace(" ", "T")

    # ------------------------------------------------------------------------------------------ #
    def set_until_time(self, timestring):
        self.until = timestring

    # ------------------------------------------------------------------------------------------ #
    def set_playlist_store(self, path):
        self.playlistdir = path

    # ------------------------------------------------------------------------------------------ #
    def get_duration(self, start, end):
        return self.__calc_duration__(start, end)

    # ------------------------------------------------------------------------------------------ #
    def get_queue(self):
        return self.queue

    # ------------------------------------------------------------------------------------------ #
    # FIXME is get_uri() needed?
    
    # def get_uri(self):
    #     if not self.playlistdir:
    #         return False
    #     if not self.datefrom:
    #         return False
    #     if not self.__calc_date_to__():
    #         return

    #     hostname = self.get("servername")
    #     port = self.get("serviceport")
    #     date_from = self.datefrom[0:16] + ":00"
    #     date_to = self.dateto[0:16] + ":00"
    #     uri = "http://" + hostname  + ":" + port + "/playlist/" + date_from + "/" + date_to

    #     return uri

    # ------------------------------------------------------------------------------------------ #
    def run(self):
        """
        Fetch calendar data and store it in the database
        """

        try:
            fetched_schedule_data = self.calendar_fetcher.fetch()

            # if nothing is fetched, return
            if fetched_schedule_data is None:
                self.queue.put("fetching_aborted Nothing fetched")
                return

            ret_schedule = []
            #            for schedule in self.fetched_schedule_data:
            #                if "start" not in schedule:
            #                    self.logger.warning("No start of schedule given. skipping the schedule: "+str(schedule))
            #                    continue
            #                if "end" not in schedule:
            #                    self.logger.warning("No end of schedule given. skipping the schedule: "+str(schedule))
            #                    continue

            for schedule in fetched_schedule_data:
                # store the schedule
                schedule_db = self.store_schedule(schedule)

                # store playlists to play
                self.logger.warning("only storing playlist")
                self.store_playlist(schedule_db, schedule_db.playlist_id, schedule["playlist"])
#                self.store_schedule_playlist(schedule_db, schedule, "schedule_fallback", 1)
#                self.store_schedule_playlist(schedule_db, schedule, "show_fallback", 2)
#                self.store_schedule_playlist(schedule_db, schedule, "station_fallback", 3)

                ret_schedule.append(schedule_db)

            # release the mutex
            self.queue.put(ret_schedule)
        except Exception as e:
            # release the mutex
            self.logger.warning("Fetching aborted due to: %s" % str(e))
            self.queue.put("fetching_aborted " + str(e))

        # terminate the thread
        return

    # ------------------------------------------------------------------------------------------ #
#    def drop_the_future(self, time_in_the_future):
#        ScheduleEntry.drop_the_future(time_in_the_future)
#        Schedule.drop_the_future(time_in_the_future)

    # ------------------------------------------------------------------------------------------ #
    def store_schedule(self, schedule):
        schedule_db = Schedule.select_show_on_datetime(schedule["start"])
        havetoadd = False

        if not schedule_db:
            self.logger.debug("no schedule with given schedule id in database => create new")
            schedule_db = Schedule()
            havetoadd = True

        # calc duration
        duration = self.__calc_duration__(schedule["start"], schedule["end"])
        schedule["duration"] = timedelta(seconds=duration).__str__()

        schedule_db.show_id = schedule["show_id"]
        schedule_db.schedule_id = schedule["schedule_id"]
        schedule_db.schedule_start = schedule["start"]
        schedule_db.schedule_end = schedule["end"]
        schedule_db.show_name = schedule["show_name"]
        schedule_db.show_hosts = schedule["show_hosts"]
        schedule_db.is_repetition = schedule["is_repetition"]
        schedule_db.funding_category = schedule["show_fundingcategory"]
        schedule_db.languages = schedule["show_languages"]
        schedule_db.type = schedule["show_type"]
        schedule_db.category = schedule["show_categories"]
        schedule_db.topic = schedule["show_topics"]
        schedule_db.musicfocus = schedule["show_musicfocus"]

        if schedule["playlist_id"] is None:
            schedule["playlist_id"] = 1
        schedule_db.playlist_id = schedule["playlist_id"]
        schedule_db.schedule_fallback_id = schedule["schedule_fallback_id"]
        schedule_db.show_fallback_id = schedule["show_fallback_id"]
        schedule_db.station_fallback_id = schedule["station_fallback_id"]

        schedule_db.store(add=havetoadd, commit=True)

        return schedule_db

    # ------------------------------------------------------------------------------------------ #
    def store_playlist(self, schedule_db, playlist_id, fetched_playlist, fallbackplaylist_type=0):
        playlist_db = Playlist.select_playlist_for_schedule(schedule_db.schedule_start, playlist_id)
        havetoadd = False

        if not playlist_db:
            playlist_db = Playlist()
            havetoadd = True

        playlist_db.playlist_id = playlist_id
        playlist_db.schedule_start = schedule_db.schedule_start
        playlist_db.show_name = schedule_db.show_name
        playlist_db.fallback_type = fallbackplaylist_type
        playlist_db.entry_count = len(fetched_playlist["entries"])

        playlist_db.store(havetoadd, True)

        self.store_playlist_entries(playlist_db, fetched_playlist)

        return playlist_db


    def store_playlist_entries(self, playlist_db, fetched_playlist):
        entry_num = 0
        for entry in fetched_playlist["entries"]:
            playlistentry_db = PlaylistEntry.select_playlistentry_for_playlist(playlist_db.artificial_id, entry_num)
            havetoadd = False
            if not playlistentry_db:
                playlistentry_db = PlaylistEntry()
                havetoadd = True

            playlistentry_db.artificial_playlist_id = playlist_db.artificial_id
            playlistentry_db.entry_num = entry_num
            playlistentry_db.uri = entry["uri"]
            playlistentry_db.filename = entry["filename"]
            playlistentry_db.duration = entry["file"]["duration"]
            playlistentry_db.store(havetoadd, True)

            self.store_playlist_entry_metadata(playlistentry_db, entry["file"]["metadata"])

            entry_num = entry_num + 1

    def store_playlist_entry_metadata(self, playlistentry_db, metadata):
        playlistentrymetadata_db = PlaylistEntryMetaData.select_metadata_for_entry(playlistentry_db.artificial_id)
        havetoadd = False
        if not playlistentrymetadata_db:
            playlistentrymetadata_db = PlaylistEntryMetaData()
            havetoadd = True

        playlistentrymetadata_db.artificial_entry_id = playlistentry_db.artificial_id
        if "artist" not in metadata:
            self.logger.warning("Artist not found in metadata for track '%s'. Setting to 'N/a'" % playlistentry_db.filename)
            playlistentrymetadata_db.artist = "N/a"
        else:
            playlistentrymetadata_db.artist = metadata["artist"]
        playlistentrymetadata_db.title = metadata["title"]
        if "album" in metadata:
            playlistentrymetadata_db.album = metadata["album"]

        playlistentrymetadata_db.store(havetoadd, True)

    # ------------------------------------------------------------------------------------------ #
    def store_playlist_entry(self, schedule_db, playlist, entry, lastentry, entrynum, fallbackplaylist_type=0):
        schedule_entry_db = Playlist.select_one().select_one_playlist_entry_for_show(schedule_db.schedule_id, fallbackplaylist_type, entrynum)
        havetoadd = False

        if not schedule_entry_db:
            self.logger.debug("no scheduleentry with id " + str(playlist["id"]) + " and pos " + str(entrynum) + " in database => creating a new one")
            # FIXME Needed? No active class declaration
            #schedule_entry_db = ScheduleEntry()
            havetoadd = True


        schedule_entry_db.playlist_id = playlist["id"]
        schedule_entry_db.entry_num = entrynum
        schedule_entry_db.schedule_id = schedule_db.schedule_id
        schedule_entry_db.uri = entry["uri"]
        schedule_entry_db.fallback_type = fallbackplaylist_type
        schedule_entry_db.entry_start = schedule_db.schedule_start + timedelta(seconds=self.get_length(lastentry))

        schedule_entry_db.calc_unix_times()
        if havetoadd:
            schedule_entry_db.define_clean_source()

        self.logger.debug("Storing entries... playlist_id: " + str(playlist["id"]) + " schedule_id: " + str(schedule_db.schedule_id) + " num: " + str(entrynum))

        schedule_entry_db.store(add=havetoadd, commit=True)

        return schedule_entry_db

    # ------------------------------------------------------------------------------------------ #
    def __calc_date_to__(self):
        if self.dateto:
            return True
        if not self.until:
            return False
        if not self.datefrom:
            return False

        date_start = datetime.strptime(self.datefrom.replace("T"," "), "%Y-%m-%d %H:%M:%S")
        time_start = date_start.strftime("%H:%M")
        day_offset = 1 if (time_start > self.until) else 0
        end_date = date_start + timedelta(day_offset)
        self.dateto = end_date.strftime("%F") + "T" + self.until

        return True

    # ------------------------------------------------------------------------------------------ #
    @staticmethod
    def __calc_duration__(start, end):
        """
        Berechnet Zeit in Sekunden aus Differenz zwischen Start und Enddatum
        @type start: datetime
        @param start: Startzeit
        @type end: datetime
        @param end: Endzeit
        @rtype:       int
        @return:      Zeit in Sekunden
        """
        sec1 = int(datetime.strptime(start[0:16].replace(" ","T"),"%Y-%m-%dT%H:%M").strftime("%s"))
        sec2 = int(datetime.strptime(end[0:16].replace(" ","T"),"%Y-%m-%dT%H:%M").strftime("%s"))
        return (sec2 - sec1)

    # ------------------------------------------------------------------------------------------ #
    def get_length(self, entry):
        if entry is None or entry.type == ScheduleEntryType.STREAM or entry.type == ScheduleEntryType.LIVE_0 or entry.type == ScheduleEntryType.LIVE_1 or entry.type == ScheduleEntryType.LIVE_2 or entry.type == ScheduleEntryType.LIVE_3 or entry.type == ScheduleEntryType.LIVE_4:
            return 0

        audio_file = FLAC(entry.cleansource)
        return audio_file.info.length

    # ------------------------------------------------------------------------------------------ #
    def __set_url__(self, type):
        url = self.config.get(type+"url")
        pos = url.find("?")

        if pos > 0:
            self.url[type] = url[0:pos]
            self.data[type] = url[pos:]
        else:
            self.url[type] = url

    # ------------------------------------------------------------------------------------------ #
    def stop(self):
        self._stop_event.set()
